/**
 * Copyright (C) 2005-2011 by Rivello Multimedia Consulting (RMC).                    
 * code [at] RivelloMultimediaConsulting [dot] com                                                  
 *                                                                      
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the      
 * "Software"), to deal in the Software without restriction, including  
 * without limitation the rights to use, copy, modify, merge, publish,  
 * distribute, sublicense, and#or sell copies of the Software, and to   
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:                                            
 *                                                                      
 * The above copyright notice and this permission notice shall be       
 * included in all copies or substantial portions of the Software.      
 *                                                                      
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,      
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF   
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR    
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.                                      
 */
// Marks the right margin of code *******************************************************************
package com.rmc.projects.pbebox2dradialgravity.components
{
	
	// --------------------------------------
	// Imports
	// --------------------------------------
	import Box2D.Collision.Shapes.b2Shape;
	import Box2D.Collision.b2Distance;
	import Box2D.Common.Math.b2Math;
	import Box2D.Common.Math.b2Vec2;
	import Box2D.Dynamics.b2Body;
	
	import com.pblabs.box2D.Box2DManagerComponent;
	import com.pblabs.box2D.Box2DSpatialComponent;
	import com.pblabs.box2D.CollisionEvent;
	import com.pblabs.engine.PBE;
	import com.pblabs.engine.components.TickedComponent;
	import com.pblabs.engine.core.ITickedObject;
	import com.pblabs.engine.core.ObjectType;
	import com.pblabs.engine.entity.IEntity;
	import com.pblabs.engine.entity.PropertyReference;
	import com.rmc.errors.SwitchStatementDefaultError;
	import com.rmc.utils.Random;
	
	import flash.events.KeyboardEvent;
	import flash.geom.Point;
	
	import flashx.textLayout.elements.BreakElement;
	
	
	// --------------------------------------
	// Metadata
	// --------------------------------------
	
	// --------------------------------------
	// Class
	// --------------------------------------
	/**
	 * <p>This is designed to...</p>
	 * 
	 * <p>AUTHOR  		: Samuel Asher Rivello</p>
	 * <p>COMPANY 		: Rivello Multimedia Consulting</p>
	 * <p>CREATION DATE 	: Jun 18, 2011</p>
	 * 
	 * @example Here is a code example. 
	 *  
	 * <listing version="3.0">
	 * <p>Describe this code example.</p>
	 *
	 * </listing>
	 *
	 * @see com.rmc.templates.TemplateClass
	 *
	 */
	public class RadialGravityComponent extends TickedComponent implements ITickedObject
	{
		
		// --------------------------------------
		// Properties
		// --------------------------------------
		// PUBLIC GETTER/SETTERS
		
		/**
		 * This is how strong the affect is on the '' (may be negative (repel), positive (attract) or 0 (no effect, default)
		 * 
		 */
		public function get radialGravityMagnitude() 						: Number 		{ return _radialGravityMagnitude_num; }
		public function set radialGravityMagnitude(aValue : Number) 		: void 			{ 
			_radialGravityMagnitude_num = aValue;
			_updateRegisterForTicks();
		}
		private var _radialGravityMagnitude_num : Number = 0;
		
		/**
		 * Slow down linear movement a little per frame (if non 0)
		 * 
		 */
		public function get linearDamping() 						: Number 		{ return _linearDamping_num; }
		public function set linearDamping(aValue : Number) 		: void 			{ 
			_linearDamping_num = aValue; 
			_updateBodyPropertiesFromSetters();
		}
		private var _linearDamping_num : Number = 0;
		
		/**
		 * Slow down angular movement a little per frame (if non 0)
		 * 
		 */
		public function get angularDamping() 					: Number 		{ return _angularDamping_num; }
		public function set angularDamping(aValue : Number) 		: void 			{ 
			_angularDamping_num = aValue; 
			_updateBodyPropertiesFromSetters();
		}
		private var _angularDamping_num : Number = 0;
		
		/**
		 * Describe this member.
		 * 
		 */
		public function get radialGravityAffectsTypes() 						: ObjectType 		{ return _radialGravityAffectsTypes; }
		public function set radialGravityAffectsTypes(aValue : ObjectType) 		: void 			{ 
			_radialGravityAffectsTypes = aValue; 
			_updateRegisterForTicks();
		}
		private var _radialGravityAffectsTypes : ObjectType;
		
		/**
		 * Describe this member.
		 * 
		 */
		public function get radialGravityType() 						: ObjectType 		{ return _radialGravityType; }
		public function set radialGravityType(aValue : ObjectType) 		: void 			{ 
			_radialGravityType = aValue; 
			_updateRegisterForTicks();
		}
		private var _radialGravityType : ObjectType;
		
		// PUBLIC CONST
		
		// PRIVATE
		/**
		 *	The Box2D Body Reference
		 */
		/**
		 * Describe this member.
		 * 
		 */
		public function get body() 						: b2Body 		{ return _body; }
		private var _body : b2Body;
		
		/**
		 *	The Box2D Manager Reference
		 */
		private var _box2DManagerComponent:Box2DManagerComponent;
		
		// --------------------------------------
		// Constructor
		// --------------------------------------
		
		/**
		 * This is the constructor.
		 * 
		 */
		public function RadialGravityComponent()
		{
			// SUPER
			super();
			
			// EVENTS
			
			// VARIABLES
			
			// PROPERTIES
			radialGravityMagnitude = 0; //USE SETTER. THIS TURNS 'OFF' GRAVITY, UNLESS XML OVERRIDES THIS VALUE
			
			//PROPERTY REFERENCES
			
			// METHODS
			
			
		}
		
		
		// --------------------------------------
		// Methods
		// --------------------------------------
		// PUBLIC
		
		//PROTECTED
		
		// --------------------------------------
		// Event Handlers
		// --------------------------------------
		/**
		 * Handles the event: Add 
		 * 
		 * @return void
		 */
		protected override function onAdd() : void
		{
			super.onAdd();
			//
			_onSetup();
		}
		
		/**
		 * Handles the event: Reset 
		 * 
		 * @return void
		 */
		protected override function onReset () : void
		{
			super.onReset();
			
			//
			_onSetup();
		}
		
		/**
		 * Handles the event: Remove 
		 * 
		 * @return void
		 */
		protected override function onRemove() : void
		{
			super.onRemove();
		}
		
		/**
		 * Handles the aEvent: <code>Tick</code>.
		 * 
		 * @param aDeltaTime_num  <code>Number</code> The delta time.
		 *  
		 * @return void
		 * 
		 */
		override public function onTick (aDeltaTime_num : Number ) : void
		{
			/*
			TODO
			1. use affecttypes and types
			2. google using mass for this type of stuff and use mass properly
			3. google that game that works off of this, and think about coding any changes to generalize it
			4. Optimize. No 'onTick' unless needed? orr....... maybe all should run code but mass-wise the earth shouldn't happen to move
			5. publish and ask for feedback.
			
			
			
			*/
			
			trace ("=======");
			if (radialGravityType) {
				trace ("affectType : " + radialGravityType.typeNames);
			}
			if (radialGravityAffectsTypes) {
				trace ("affectsWithTypes : " + radialGravityAffectsTypes.typeNames);
			}
			
			//LOOP THROUGH ALL BODIES, IF WE WANT TO AFFECT THEM WITH GRAVITY, DO SO...
			var next_b2Body : b2Body = _box2DManagerComponent.world.GetBodyList();
			var next_radialGravityComponent : RadialGravityComponent;
			
			//LOOP THROUGH *ALL* BODIES AND IF ITS TYPE IS WHAT I WANT TO AFFECT, THEN AFFECT IT.
			for (next_b2Body; next_b2Body; next_b2Body = next_b2Body.GetNext()) {
				
				//DOES MY TYPE MATCH ITS TYPE
				next_radialGravityComponent = _getRadialGravityComponentForB2Body (next_b2Body);
				if (next_radialGravityComponent != null && PBE.objectTypeManager.doesTypeOverlap(radialGravityAffectsTypes, next_radialGravityComponent.radialGravityType.typeName)) {
					
					trace ("found : " + radialGravityAffectsTypes.typeName + " and " + next_radialGravityComponent.radialGravityType.typeName);
					//IS IT *NOT ME* (WE DON'T WANT TO AFFECT OURSELVES)
					if (body != next_radialGravityComponent.body) {
						
						//ADD FORCE OF RADIAL GRAVITY
						var force_b2vec2 : b2Vec2 = b2Math.SubtractVV(next_b2Body.GetPosition(), _body.GetPosition());
						force_b2vec2.Normalize();
						force_b2vec2.Multiply( (_body.GetMass() - next_b2Body.GetMass() )*radialGravityMagnitude/100	);
						next_b2Body.ApplyImpulse (force_b2vec2.Negative(), next_b2Body.GetPosition() );
						
						
						//trace ("is f: " + force_b2vec2 + " : d" + d + " --- " + "(" +next_b2Body.GetPosition().x + " , " + next_b2Body.GetPosition().y + ")" + " and " + "(" + _b2Body.GetPosition().x + " , " + _b2Body.GetPosition().y + ")");
						//trace ("f1 " + force_b2vec2.x + " , " + force_b2vec2.y);
						//force_b2vec2.Normalize();
						//trace ("f2 " + force_b2vec2.x + " , " + force_b2vec2.y);
						//next_b2Body.ApplyForce (force_b2vec2.Negative(), next_b2Body.GetPosition() );
					} 
				}
			}
		}
		
		/*
		(id) init {
		//... Add at end of init after body->CreateFixture(&fixtureDef);
		// Create our static "Planet"
		b2CircleShape shape;
		shape.m_radius = 1.0f;
		shape.m_p.Set(8.0f, 8.0f);
		b2FixtureDef fd;
		fd.shape = &shape;
		planet = groundBody->CreateFixture(&fd);
		}
		
		-(void) tick: (ccTime) dt
		{
		int32 velocityIterations = 8;
		int32 positionIterations = 1;
		
		world->Step(dt, velocityIterations, positionIterations);
		
		for (b2Body* b = world->GetBodyList(); b; b = b->GetNext())
		{
		b2Body* ground = planet->GetBody();
		b2CircleShape* circle = (b2CircleShape*)planet->GetShape();
		// Get position of our "Planet"
		b2Vec2 center = ground->GetWorldPoint(circle->m_p);
		// Get position of our current body in the iteration
		b2Vec2 position = b->GetPosition();
		// Get the distance between the two objects.
		b2Vec2 d = center - position;
		// The further away the objects are, the weaker the gravitational force is
		float force = 250.0f / d.LengthSquared(); // 150 can be changed to adjust the amount of force
		d.Normalize();
		b2Vec2 F = force * d;
		// Finally apply a force on the body in the direction of the "Planet"
		b->ApplyForce(F, position);
		
		if (b->GetUserData() != NULL) {
		CCSprite *myActor = (CCSprite*)b->GetUserData();
		myActor.position = CGPointMake( b->GetPosition().x * PTM_RATIO, b->GetPosition().y * PTM_RATIO);
		myActor.rotation = -1 * CC_RADIANS_TO_DEGREES(b->GetAngle());
		}	
		}
		}
		*/
		
		//	PRIVATE
		/**
		 * Called both on onReset and onAdd
		 * 
		 * @return void
		 */
		private function _onSetup () : void
		{
			//BOX2D
			_box2DManagerComponent  = PBE.lookupComponentByName("SceneEntity", "Box2DManagerComponent") as Box2DManagerComponent;
			//
			var box2DSpatialComponent : Box2DSpatialComponent = (owner.lookupComponentByName("Spatial") as Box2DSpatialComponent);
			_body  = box2DSpatialComponent.body;
			_updateBodyPropertiesFromSetters();
			
		}
		
		/**
		 * Update if we are registered for ticking and thus calculating gravity *ON OTHERS* bodies.
		 * 
		 * 	NOTE. 	Its common for this component to hold valuable values, and NOT be calculating gravity,
		 * 			but instead for other bodies to be calculating gravity *ON THIS*. 
		 * 
		 * @return void
		 * 
		 */
		private function _updateRegisterForTicks () : void
		{
			if (radialGravityMagnitude != 0 && 
				radialGravityAffectsTypes != null && 
				radialGravityAffectsTypes.typeNames.length > 0) {
				
				registerForTicks = true;
			} else {
				registerForTicks = false;
			}
			
		}
		
		/**
		 * Update Body based on component setters. 
		 * 
		 * @return void
		 * 
		 */
		private function _updateBodyPropertiesFromSetters () : void
		{
			if (_body) {
				_body.m_linearDamping 		= _linearDamping_num;
				_body.m_angularDamping 	= _angularDamping_num;
			}
			
		}
		/**
		 * Lookup the RadialGravityComponent for a specific b2Body
		 * 
		 * 	TODO: Is there a better way? Is 'entity' stored within the b2Body or
		 *  is there a lookup method that can help anywhere?
		 * 
		 * @return RadialGravityComponent
		 * 
		 */
		private function _getRadialGravityComponentForB2Body (aB2Body : b2Body) : RadialGravityComponent
		{
			//trace ("\n-----");
			var radialGravityComponent : RadialGravityComponent;
			for each (var object : Object in PBE.nameManager.objectList) {
				if (object is IEntity) {
					radialGravityComponent  = (object as IEntity).lookupComponentByName("RadialGravityComponent") as RadialGravityComponent;
					if ( radialGravityComponent) {
						if ( radialGravityComponent.body == aB2Body) {
							//trace ("n: " + radialGravityComponent.body);
							return radialGravityComponent;
						}
					}
				}
			}
			
			return null;
		}
		
		
	}
}
